/*
 * DAC Dithering angelehnt an AppNote R01AN3795EU0100 "RL78 Family" "RL78 8-bit DAC Dithering"
 * Es werden 16 DAC-Ausgabewerte ermittelt, die zyklisch an den DAC übergeben werden müssen - HW Befilterung nutürlich vorausgesetzt
 *
 * Alles ist in Fixpoint (ohne Floating Datentypen) umgesetzt und damit ggf. auch für 
 * einfache MCUs (8/16/32 Bit ohne FPU) geeignet - ausreichend RAM vorausgesetzt!
 */

#include <stdio.h>              /* Wird nur für die Demo benötigt (printf Ausgabe)*/
#include <math.h>               /* Wird nur für die Demo benötigt (float_t Datentyp) */

#include <stdint.h>             /* Wird für die uintX_t Datentypen benötigt */
#include <stdbool.h>            /* wird für den bool Datentyp benötigt */

#include "dithering.h"          /* Hier geht's zur Magie :-P */
                                /* Auch wird hier die HW-Auflösung (also vor Dithering) und Referenzspannung festgelegt */

/* Nur als Debug Ausgabe gedacht - schneller "Dump" des Werte-Arrays 
 * Per get_dithering_dac_value(0..15) muss man die 16 DAC werte abfragen 
 * und an seinen DAC senden...
 */
void dump_dithering_table(void);
void dump_dithering_table(void){
    uint8_t loop_counter = 0u;
    printf("\nDithering table:");
    for(loop_counter = 0u; loop_counter < 16u; loop_counter++){
        printf("\n\t%u, \t%u", loop_counter, get_dithering_dac_value(loop_counter));
    }
}

/* "Simuliert" die Ausgabe des DACs mit Dithering. 
 * Natürlich sind hierzu externe Filter in HW nötig!
 * Dient nur dem Test - sollte nicht auf MCU-HW portiert werden!
 */
uint32_t dac_simulated_output(void){
    uint32_t average = 0u;
    uint32_t total = 0u;
    uint32_t u_dac_out = 0;
    uint8_t loop_counter = 0u;

    for(loop_counter = 0; loop_counter < 16u; loop_counter++){
        total = (total + (get_dithering_dac_value(loop_counter) * 1000));
    }
    average = total / 16;
    u_dac_out = (average * DAC_1_STEP_IN_MICROVOLT)/1000;
    return u_dac_out;
}

/* Berechnet die Abweichung zwischen SOLL und IST des DACs
 * Dient nur dem Test - sollte nicht auf MCU-HW portiert werden!
 */
double_t deviation(uint32_t exact, uint32_t approxinate){
    double_t exa = 0.0f;
    double_t approx = 0.0f;

    exa = (double_t)exact;
    approx = (double_t)approxinate;

    return (((approx-exa)/exa)*100.0f);
}

/* Kurzes Verwendungsbeispiel in form eines statischen Konsolenprogramms */
int main(void){
    uint32_t loc_vout_in_microvolt = 0u;    /* die Ausgangsspannung die der DAC ausgeben soll - wird später gesetzt*/  
    uint32_t loc_dac_step_base = 0u;        /* der rohe DAC-Zahlenwert - wird später berechnet */
    uint8_t loc_dithering_sequence = 0u;    /* welche Dithering-Sequenez benutzt wird - wird später berechnet */

    /* Die DAC Ausgangsspannung auf 1,63123V setzen - wie sinnvoll das ist sei erst einmal ausßer acht gelassen */
    loc_vout_in_microvolt = 1631230u;       /* Die Spannung die der DAC ausgeben soll in µV (wegen Fixpoint!) */  
 
    /* Vorsichtshalber erst einmal die Dithering-Tabelle "bereinigen" */
    clear_dithering_table();

    /* Ausgabe der aktuellen Konfiguration (per #define) */
    printf("\nSelceted DAC-res: \t%u (Bits)",DAC_RES_IN_BITS);
    printf("\nSelceted Vref: \t\t%u (milliVolt)",VREF_IN_MILLIVOLT);
    printf("\nSelected DAC-Out: \t%u (microVolt)", loc_vout_in_microvolt);

    /* Rohen DAC Zahlenwert ermitteln */
    loc_dac_step_base = calulate_dac_base_steps(loc_vout_in_microvolt);
    printf("\n= DAC-STEPS:\t\t%u", DAC_MAX_STEPS);
    printf("\n= DAC-Base value: \t%u", loc_dac_step_base);
    

    /* Dithering Sequenz ermitteln */
    loc_dithering_sequence = get_dither_sequence();
    printf("\n= Dithering seq.: \t%u", loc_dithering_sequence);

    fill_dac_value_table(loc_dac_step_base, loc_dithering_sequence);
    
    /* Anzeige nur zum Debuggen */
    //dump_dithering_table();

    printf("\n= DAC-output: \t\t%u (microVolt)", dac_simulated_output());
    printf("\n= Deviation: \t\t%lf%%", deviation(loc_vout_in_microvolt, dac_simulated_output()));

    return 0;
}